"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import shutil
import tempfile

import salt.states.virt as virt
import salt.utils.files
from salt.exceptions import SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class LibvirtMock(MagicMock):  # pylint: disable=too-many-ancestors
    """
    libvirt library mockup
    """

    class libvirtError(Exception):  # pylint: disable=invalid-name
        """
        libvirt error mockup
        """

        def get_error_message(self):
            """
            Fake function return error message
            """
            return str(self)


class LibvirtTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.libvirt
    """

    def setup_loader_modules(self):
        self.mock_libvirt = (
            LibvirtMock()
        )  # pylint: disable=attribute-defined-outside-init
        self.addCleanup(delattr, self, "mock_libvirt")
        loader_globals = {"libvirt": self.mock_libvirt}
        return {virt: loader_globals}

    @classmethod
    def setUpClass(cls):
        cls.pki_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.pki_dir)
        del cls.pki_dir

    # 'keys' function tests: 1

    def test_keys(self):
        """
        Test to manage libvirt keys.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            name = "sunrise"

            ret = {"name": name, "result": True, "comment": "", "changes": {}}

            mock = MagicMock(
                side_effect=[
                    [],
                    ["libvirt.servercert.pem"],
                    {"libvirt.servercert.pem": "A"},
                ]
            )
            with patch.dict(virt.__salt__, {"pillar.ext": mock}):
                comt = "All keys are correct"
                ret.update({"comment": comt})
                self.assertDictEqual(virt.keys(name, basepath=self.pki_dir), ret)

                with patch.dict(virt.__opts__, {"test": True}):
                    comt = "Libvirt keys are set to be updated"
                    ret.update({"comment": comt, "result": None})
                    self.assertDictEqual(virt.keys(name, basepath=self.pki_dir), ret)

                with patch.dict(virt.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Updated libvirt certs and keys"
                        ret.update(
                            {
                                "comment": comt,
                                "result": True,
                                "changes": {"servercert": "new"},
                            }
                        )
                        self.assertDictEqual(
                            virt.keys(name, basepath=self.pki_dir), ret
                        )

    def test_keys_with_expiration_days(self):
        """
        Test to manage libvirt keys.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            name = "sunrise"

            ret = {"name": name, "result": True, "comment": "", "changes": {}}

            mock = MagicMock(
                side_effect=[
                    [],
                    ["libvirt.servercert.pem"],
                    {"libvirt.servercert.pem": "A"},
                ]
            )
            with patch.dict(virt.__salt__, {"pillar.ext": mock}):
                comt = "All keys are correct"
                ret.update({"comment": comt})
                self.assertDictEqual(
                    virt.keys(name, basepath=self.pki_dir, expiration_days=700), ret
                )

                with patch.dict(virt.__opts__, {"test": True}):
                    comt = "Libvirt keys are set to be updated"
                    ret.update({"comment": comt, "result": None})
                    self.assertDictEqual(
                        virt.keys(name, basepath=self.pki_dir, expiration_days=700), ret
                    )

                with patch.dict(virt.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Updated libvirt certs and keys"
                        ret.update(
                            {
                                "comment": comt,
                                "result": True,
                                "changes": {"servercert": "new"},
                            }
                        )
                        self.assertDictEqual(
                            virt.keys(name, basepath=self.pki_dir, expiration_days=700),
                            ret,
                        )

    def test_keys_with_state(self):
        """
        Test to manage libvirt keys.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            name = "sunrise"

            ret = {"name": name, "result": True, "comment": "", "changes": {}}

            mock = MagicMock(
                side_effect=[
                    [],
                    ["libvirt.servercert.pem"],
                    {"libvirt.servercert.pem": "A"},
                ]
            )
            with patch.dict(virt.__salt__, {"pillar.ext": mock}):
                comt = "All keys are correct"
                ret.update({"comment": comt})
                self.assertDictEqual(
                    virt.keys(name, basepath=self.pki_dir, st="California"), ret
                )

                with patch.dict(virt.__opts__, {"test": True}):
                    comt = "Libvirt keys are set to be updated"
                    ret.update({"comment": comt, "result": None})
                    self.assertDictEqual(
                        virt.keys(name, basepath=self.pki_dir, st="California"), ret
                    )

                with patch.dict(virt.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Updated libvirt certs and keys"
                        ret.update(
                            {
                                "comment": comt,
                                "result": True,
                                "changes": {"servercert": "new"},
                            }
                        )
                        self.assertDictEqual(
                            virt.keys(name, basepath=self.pki_dir, st="California"), ret
                        )

    def test_keys_with_all_options(self):
        """
        Test to manage libvirt keys.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            name = "sunrise"

            ret = {"name": name, "result": True, "comment": "", "changes": {}}

            mock = MagicMock(
                side_effect=[
                    [],
                    ["libvirt.servercert.pem"],
                    {"libvirt.servercert.pem": "A"},
                ]
            )
            with patch.dict(virt.__salt__, {"pillar.ext": mock}):
                comt = "All keys are correct"
                ret.update({"comment": comt})
                self.assertDictEqual(
                    virt.keys(
                        name,
                        basepath=self.pki_dir,
                        country="USA",
                        st="California",
                        locality="Los_Angeles",
                        organization="SaltStack",
                        expiration_days=700,
                    ),
                    ret,
                )

                with patch.dict(virt.__opts__, {"test": True}):
                    comt = "Libvirt keys are set to be updated"
                    ret.update({"comment": comt, "result": None})
                    self.assertDictEqual(
                        virt.keys(
                            name,
                            basepath=self.pki_dir,
                            country="USA",
                            st="California",
                            locality="Los_Angeles",
                            organization="SaltStack",
                            expiration_days=700,
                        ),
                        ret,
                    )

                with patch.dict(virt.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Updated libvirt certs and keys"
                        ret.update(
                            {
                                "comment": comt,
                                "result": True,
                                "changes": {"servercert": "new"},
                            }
                        )
                        self.assertDictEqual(
                            virt.keys(
                                name,
                                basepath=self.pki_dir,
                                country="USA",
                                st="California",
                                locality="Los_Angeles",
                                organization="SaltStack",
                                expiration_days=700,
                            ),
                            ret,
                        )

    def test_pool_defined(self):
        """
        pool_defined state test cases.
        """
        ret = {"name": "mypool", "changes": {}, "result": True, "comment": ""}
        mocks = {
            mock: MagicMock(return_value=True)
            for mock in ["define", "autostart", "build"]
        }
        with patch.dict(virt.__opts__, {"test": False}):
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        side_effect=[
                            {},
                            {"mypool": {"state": "stopped", "autostart": True}},
                        ]
                    ),
                    "virt.pool_define": mocks["define"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_set_autostart": mocks["autostart"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool defined, marked for autostart"},
                        "comment": "Pool mypool defined, marked for autostart",
                    }
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                        transient=True,
                        autostart=True,
                        connection="myconnection",
                        username="user",
                        password="secret",
                    ),
                    ret,
                )
                mocks["define"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    transient=True,
                    start=False,
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                mocks["autostart"].assert_called_with(
                    "mypool",
                    state="on",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                mocks["build"].assert_called_with(
                    "mypool",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )

            # Define a pool that doesn't handle build
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        side_effect=[
                            {},
                            {"mypool": {"state": "stopped", "autostart": True}},
                        ]
                    ),
                    "virt.pool_define": mocks["define"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_set_autostart": mocks["autostart"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool defined, marked for autostart"},
                        "comment": "Pool mypool defined, marked for autostart",
                    }
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="rbd",
                        source={
                            "name": "libvirt-pool",
                            "hosts": ["ses2.tf.local", "ses3.tf.local"],
                            "auth": {
                                "username": "libvirt",
                                "password": "AQAz+PRdtquBBRAASMv7nlMZYfxIyLw3St65Xw==",
                            },
                        },
                        autostart=True,
                    ),
                    ret,
                )
                mocks["define"].assert_called_with(
                    "mypool",
                    ptype="rbd",
                    target=None,
                    permissions=None,
                    source_devices=None,
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=["ses2.tf.local", "ses3.tf.local"],
                    source_auth={
                        "username": "libvirt",
                        "password": "AQAz+PRdtquBBRAASMv7nlMZYfxIyLw3St65Xw==",
                    },
                    source_name="libvirt-pool",
                    source_format=None,
                    source_initiator=None,
                    start=False,
                    transient=False,
                    connection=None,
                    username=None,
                    password=None,
                )
                mocks["autostart"].assert_called_with(
                    "mypool",
                    state="on",
                    connection=None,
                    username=None,
                    password=None,
                )
                mocks["build"].assert_not_called()

            mocks["update"] = MagicMock(return_value=False)
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "stopped", "autostart": True}}
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_build": mocks["build"],
                },
            ):
                ret.update({"changes": {}, "comment": "Pool mypool unchanged"})
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["build"].assert_not_called()

            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(return_value={}),
                    "virt.pool_define": MagicMock(
                        side_effect=self.mock_libvirt.libvirtError("Some error")
                    ),
                },
            ):
                ret.update({"changes": {}, "comment": "Some error", "result": False})
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            # Test case with update and autostart change on stopped pool
            for mock in mocks:
                mocks[mock].reset_mock()
            mocks["update"] = MagicMock(return_value=True)
            mocks["build"] = MagicMock(
                side_effect=self.mock_libvirt.libvirtError("Existing VG")
            )
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "stopped", "autostart": True}}
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_set_autostart": mocks["autostart"],
                    "virt.pool_build": mocks["build"],
                },
            ):
                ret.update(
                    {
                        "changes": {
                            "mypool": "Pool updated, built, autostart flag changed"
                        },
                        "comment": "Pool mypool updated, built, autostart flag changed",
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        autostart=False,
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["build"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["autostart"].assert_called_with(
                    "mypool", state="off", connection=None, username=None, password=None
                )
                mocks["update"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    connection=None,
                    username=None,
                    password=None,
                )

            # test case with update and no autostart change on running pool
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={
                            "mypool": {"state": "running", "autostart": False}
                        }
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_build": mocks["build"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool updated"},
                        "comment": "Pool mypool updated",
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        autostart=False,
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["build"].assert_not_called()
                mocks["update"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    connection=None,
                    username=None,
                    password=None,
                )

        with patch.dict(virt.__opts__, {"test": True}):
            # test case with test=True and no change
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "running", "autostart": True}}
                    ),
                    "virt.pool_update": MagicMock(return_value=False),
                },
            ):
                ret.update(
                    {"changes": {}, "comment": "Pool mypool unchanged", "result": True}
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            # test case with test=True and pool to be defined
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(return_value={}),
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool defined, marked for autostart"},
                        "comment": "Pool mypool defined, marked for autostart",
                        "result": None,
                    }
                )
                self.assertDictEqual(
                    virt.pool_defined(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                        transient=True,
                        autostart=True,
                        connection="myconnection",
                        username="user",
                        password="secret",
                    ),
                    ret,
                )

    def test_pool_running(self):
        """
        pool_running state test cases.
        """
        ret = {"name": "mypool", "changes": {}, "result": True, "comment": ""}
        mocks = {
            mock: MagicMock(return_value=True)
            for mock in ["define", "autostart", "build", "start", "stop"]
        }
        with patch.dict(virt.__opts__, {"test": False}):
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        side_effect=[
                            {},
                            {"mypool": {"state": "stopped", "autostart": True}},
                        ]
                    ),
                    "virt.pool_define": mocks["define"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_start": mocks["start"],
                    "virt.pool_set_autostart": mocks["autostart"],
                },
            ):
                ret.update(
                    {
                        "changes": {
                            "mypool": "Pool defined, marked for autostart, started"
                        },
                        "comment": "Pool mypool defined, marked for autostart, started",
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                        transient=True,
                        autostart=True,
                        connection="myconnection",
                        username="user",
                        password="secret",
                    ),
                    ret,
                )
                mocks["define"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    transient=True,
                    start=False,
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                mocks["autostart"].assert_called_with(
                    "mypool",
                    state="on",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                mocks["build"].assert_called_with(
                    "mypool",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )
                mocks["start"].assert_called_with(
                    "mypool",
                    connection="myconnection",
                    username="user",
                    password="secret",
                )

            mocks["update"] = MagicMock(return_value=False)
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "running", "autostart": True}}
                    ),
                    "virt.pool_update": MagicMock(return_value=False),
                },
            ):
                ret.update({"changes": {}, "comment": "Pool mypool already running"})
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "stopped", "autostart": True}}
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_start": mocks["start"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool started"},
                        "comment": "Pool mypool started",
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["start"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["build"].assert_not_called()

            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(return_value={}),
                    "virt.pool_define": MagicMock(
                        side_effect=self.mock_libvirt.libvirtError("Some error")
                    ),
                },
            ):
                ret.update({"changes": {}, "comment": "Some error", "result": False})
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            # Test case with update and autostart change on stopped pool
            for mock in mocks:
                mocks[mock].reset_mock()
            mocks["update"] = MagicMock(return_value=True)
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "stopped", "autostart": True}}
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_set_autostart": mocks["autostart"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_start": mocks["start"],
                },
            ):
                ret.update(
                    {
                        "changes": {
                            "mypool": (
                                "Pool updated, built, autostart flag changed, started"
                            )
                        },
                        "comment": (
                            "Pool mypool updated, built, autostart flag changed,"
                            " started"
                        ),
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        autostart=False,
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["start"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["build"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["autostart"].assert_called_with(
                    "mypool", state="off", connection=None, username=None, password=None
                )
                mocks["update"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    connection=None,
                    username=None,
                    password=None,
                )

            # test case with update and no autostart change on running pool
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={
                            "mypool": {"state": "running", "autostart": False}
                        }
                    ),
                    "virt.pool_update": mocks["update"],
                    "virt.pool_build": mocks["build"],
                    "virt.pool_start": mocks["start"],
                    "virt.pool_stop": mocks["stop"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool updated, restarted"},
                        "comment": "Pool mypool updated, restarted",
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        autostart=False,
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )
                mocks["stop"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["start"].assert_called_with(
                    "mypool", connection=None, username=None, password=None
                )
                mocks["build"].assert_not_called()
                mocks["update"].assert_called_with(
                    "mypool",
                    ptype="logical",
                    target="/dev/base",
                    permissions={
                        "mode": "0770",
                        "owner": 1000,
                        "group": 100,
                        "label": "seclabel",
                    },
                    source_devices=[{"path": "/dev/sda"}],
                    source_dir=None,
                    source_adapter=None,
                    source_hosts=None,
                    source_auth=None,
                    source_name=None,
                    source_format=None,
                    source_initiator=None,
                    connection=None,
                    username=None,
                    password=None,
                )

        with patch.dict(virt.__opts__, {"test": True}):
            # test case with test=True and no change
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "running", "autostart": True}}
                    ),
                    "virt.pool_update": MagicMock(return_value=False),
                },
            ):
                ret.update(
                    {
                        "changes": {},
                        "comment": "Pool mypool already running",
                        "result": True,
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            # test case with test=True and started
            for mock in mocks:
                mocks[mock].reset_mock()
            mocks["update"] = MagicMock(return_value=False)
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(
                        return_value={"mypool": {"state": "stopped", "autostart": True}}
                    ),
                    "virt.pool_update": mocks["update"],
                },
            ):
                ret.update(
                    {
                        "changes": {"mypool": "Pool started"},
                        "comment": "Pool mypool started",
                        "result": None,
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        source={"devices": [{"path": "/dev/sda"}]},
                    ),
                    ret,
                )

            # test case with test=True and pool to be defined
            for mock in mocks:
                mocks[mock].reset_mock()
            with patch.dict(
                virt.__salt__,
                {  # pylint: disable=no-member
                    "virt.pool_info": MagicMock(return_value={}),
                },
            ):
                ret.update(
                    {
                        "changes": {
                            "mypool": "Pool defined, marked for autostart, started"
                        },
                        "comment": "Pool mypool defined, marked for autostart, started",
                        "result": None,
                    }
                )
                self.assertDictEqual(
                    virt.pool_running(
                        "mypool",
                        ptype="logical",
                        target="/dev/base",
                        permissions={
                            "mode": "0770",
                            "owner": 1000,
                            "group": 100,
                            "label": "seclabel",
                        },
                        source={"devices": [{"path": "/dev/sda"}]},
                        transient=True,
                        autostart=True,
                        connection="myconnection",
                        username="user",
                        password="secret",
                    ),
                    ret,
                )

    def test_pool_deleted(self):
        """
        Test the pool_deleted state
        """
        # purge=False test case, stopped pool
        with patch.dict(
            virt.__salt__,
            {
                "virt.pool_info": MagicMock(
                    return_value={"test01": {"state": "stopped", "type": "dir"}}
                ),
                "virt.pool_undefine": MagicMock(return_value=True),
            },
        ):
            expected = {
                "name": "test01",
                "changes": {
                    "stopped": False,
                    "deleted_volumes": [],
                    "deleted": False,
                    "undefined": True,
                },
                "result": True,
                "comment": "",
            }

            with patch.dict(virt.__opts__, {"test": False}):
                self.assertDictEqual(expected, virt.pool_deleted("test01"))

            with patch.dict(virt.__opts__, {"test": True}):
                expected["result"] = None
                self.assertDictEqual(expected, virt.pool_deleted("test01"))

        # purge=False test case
        with patch.dict(
            virt.__salt__,
            {
                "virt.pool_info": MagicMock(
                    return_value={"test01": {"state": "running", "type": "dir"}}
                ),
                "virt.pool_undefine": MagicMock(return_value=True),
                "virt.pool_stop": MagicMock(return_value=True),
            },
        ):
            expected = {
                "name": "test01",
                "changes": {
                    "stopped": True,
                    "deleted_volumes": [],
                    "deleted": False,
                    "undefined": True,
                },
                "result": True,
                "comment": "",
            }

            with patch.dict(virt.__opts__, {"test": False}):
                self.assertDictEqual(expected, virt.pool_deleted("test01"))

            with patch.dict(virt.__opts__, {"test": True}):
                expected["result"] = None
                self.assertDictEqual(expected, virt.pool_deleted("test01"))

        # purge=True test case

        with patch.dict(
            virt.__salt__,
            {
                "virt.pool_info": MagicMock(
                    return_value={"test01": {"state": "running", "type": "dir"}}
                ),
                "virt.pool_list_volumes": MagicMock(
                    return_value=["vm01.qcow2", "vm02.qcow2"]
                ),
                "virt.pool_refresh": MagicMock(return_value=True),
                "virt.volume_delete": MagicMock(return_value=True),
                "virt.pool_stop": MagicMock(return_value=True),
                "virt.pool_delete": MagicMock(return_value=True),
                "virt.pool_undefine": MagicMock(return_value=True),
            },
        ):
            expected = {
                "name": "test01",
                "changes": {
                    "stopped": True,
                    "deleted_volumes": ["vm01.qcow2", "vm02.qcow2"],
                    "deleted": True,
                    "undefined": True,
                },
                "result": True,
                "comment": "",
            }

            with patch.dict(virt.__opts__, {"test": False}):
                self.assertDictEqual(expected, virt.pool_deleted("test01", purge=True))

            with patch.dict(virt.__opts__, {"test": True}):
                expected["result"] = None
                self.assertDictEqual(expected, virt.pool_deleted("test01", purge=True))

        # Case of backend not unsupporting delete operations
        with patch.dict(
            virt.__salt__,
            {
                "virt.pool_info": MagicMock(
                    return_value={"test01": {"state": "running", "type": "iscsi"}}
                ),
                "virt.pool_stop": MagicMock(return_value=True),
                "virt.pool_undefine": MagicMock(return_value=True),
            },
        ):
            expected = {
                "name": "test01",
                "changes": {
                    "stopped": True,
                    "deleted_volumes": [],
                    "deleted": False,
                    "undefined": True,
                },
                "result": True,
                "comment": (
                    'Unsupported actions for pool of type "iscsi": deleting volume,'
                    " deleting pool"
                ),
            }

            with patch.dict(virt.__opts__, {"test": False}):
                self.assertDictEqual(expected, virt.pool_deleted("test01", purge=True))

            with patch.dict(virt.__opts__, {"test": True}):
                expected["result"] = None
                self.assertDictEqual(expected, virt.pool_deleted("test01", purge=True))

    def test_volume_defined(self):
        """
        test the virt.volume_defined state
        """
        with patch.dict(virt.__opts__, {"test": False}):
            # test case: creating a volume
            define_mock = MagicMock()
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(return_value={"mypool": {}}),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {"mypool/myvol": {"old": "", "new": "defined"}},
                        "result": True,
                        "comment": "Volume myvol defined in pool mypool",
                    },
                )
                define_mock.assert_called_once_with(
                    "mypool",
                    "myvol",
                    "1234",
                    allocation="12345",
                    format="qcow2",
                    type="file",
                    permissions={"mode": "0755", "owner": "123", "group": "456"},
                    backing_store={"path": "/path/to/image", "format": "raw"},
                    nocow=True,
                    connection="test:///",
                    username="jdoe",
                    password="supersecret",
                )

            # test case: with existing volume
            define_mock.reset_mock()
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "1293942784",
                                    "backing_store": {
                                        "path": "/path/to/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": True,
                        "comment": "volume is existing",
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different sizes
            define_mock.reset_mock()
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "12345",
                                    "backing_store": {
                                        "path": "/path/to/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": True,
                        "comment": (
                            "The capacity of the volume is different, but no resize"
                            " performed"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different backing store
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/other/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different format
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "raw",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: no pool
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/other/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different format
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=[]),
                    "virt.volume_infos": MagicMock(return_value={}),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertRaisesRegex(
                    SaltInvocationError,
                    "Storage pool mypool not existing",
                    virt.volume_defined,
                    "mypool",
                    "myvol",
                    "1234",
                    allocation="12345",
                    format="qcow2",
                    type="file",
                    permissions={"mode": "0755", "owner": "123", "group": "456"},
                    backing_store={"path": "/path/to/image", "format": "raw"},
                    nocow=True,
                    connection="test:///",
                    username="jdoe",
                    password="supersecret",
                )

        # Test mode cases
        with patch.dict(virt.__opts__, {"test": True}):
            # test case: creating a volume
            define_mock.reset_mock()
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(return_value={"mypool": {}}),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {"mypool/myvol": {"old": "", "new": "defined"}},
                        "result": None,
                        "comment": "Volume myvol would be defined in pool mypool",
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different sizes
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "12345",
                                    "backing_store": {
                                        "path": "/path/to/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": True,
                        "comment": (
                            "The capacity of the volume is different, but no resize"
                            " performed"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different backing store
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/other/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different format
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "raw",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: no pool
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=["mypool"]),
                    "virt.volume_infos": MagicMock(
                        return_value={
                            "mypool": {
                                "myvol": {
                                    "format": "qcow2",
                                    "capacity": "1234",
                                    "backing_store": {
                                        "path": "/path/to/other/image",
                                        "format": "raw",
                                    },
                                }
                            }
                        }
                    ),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertDictEqual(
                    virt.volume_defined(
                        "mypool",
                        "myvol",
                        "1234",
                        allocation="12345",
                        format="qcow2",
                        type="file",
                        permissions={"mode": "0755", "owner": "123", "group": "456"},
                        backing_store={"path": "/path/to/image", "format": "raw"},
                        nocow=True,
                        connection="test:///",
                        username="jdoe",
                        password="supersecret",
                    ),
                    {
                        "name": "myvol",
                        "changes": {},
                        "result": False,
                        "comment": (
                            "A volume with the same name but different backing store or"
                            " format is existing"
                        ),
                    },
                )
                define_mock.assert_not_called()

            # test case: with existing volume, different format
            with patch.dict(
                virt.__salt__,
                {
                    "virt.list_pools": MagicMock(return_value=[]),
                    "virt.volume_infos": MagicMock(return_value={}),
                    "virt.volume_define": define_mock,
                },
            ):
                self.assertRaisesRegex(
                    SaltInvocationError,
                    "Storage pool mypool not existing",
                    virt.volume_defined,
                    "mypool",
                    "myvol",
                    "1234",
                    allocation="12345",
                    format="qcow2",
                    type="file",
                    permissions={"mode": "0755", "owner": "123", "group": "456"},
                    backing_store={"path": "/path/to/image", "format": "raw"},
                    nocow=True,
                    connection="test:///",
                    username="jdoe",
                    password="supersecret",
                )



"""
Salt states to create and manage VMware vSphere datacenters (datacenters).

:codeauthor: `Alexandru Bleotu <alexandru.bleotu@morganstaley.com>`

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESX data center module. Because the Salt extensions are newer
    and actively supported by VMware, they are more compatible with current
    versions of ESXi and they work well with the latest features in the VMware
    product line.


Dependencies
============

- pyVmomi Python Module

States
======

datacenter_configured
---------------------

Makes sure a datacenter exists and is correctly configured.

If the state is run by an ``esxdatacenter`` minion, the name of the datacenter
is retrieved from the proxy details, otherwise the datacenter has the same name
as the state.

Supported proxies: esxdatacenter


Example:

1. Make sure that a datacenter named ``target_dc`` exists on the vCenter, using a
``esxdatacenter`` proxy:

Proxy minion configuration (connects passthrough to the vCenter):

.. code-block:: yaml

    proxy:
      proxytype: esxdatacenter
      datacenter: target_dc
      vcenter: vcenter.fake.com
      mechanism: sspi
      domain: fake.com
      principal: host

State configuration:

.. code-block:: yaml

    datacenter_state:
      esxdatacenter.datacenter_configured
"""

import logging
from functools import wraps

import salt.exceptions

# Get Logging Started
log = logging.getLogger(__name__)
LOGIN_DETAILS = {}


def __virtual__():
    return "esxdatacenter"


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            3008,
            "The 'esxdatacenter' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3008 in favor of the "
            "saltext.vmware Salt Extension. "
            "(https://github.com/saltstack/salt-ext-modules-vmware)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def mod_init(low):
    return True


@_deprecation_message
def datacenter_configured(name):
    """
    Makes sure a datacenter exists.

    If the state is run by an ``esxdatacenter`` minion, the name of the
    datacenter is retrieved from the proxy details, otherwise the datacenter
    has the same name as the state.

    Supported proxies: esxdatacenter

    name:
        Datacenter name. Ignored if the proxytype is ``esxdatacenter``.
    """
    proxy_type = __salt__["vsphere.get_proxy_type"]()
    if proxy_type == "esxdatacenter":
        dc_name = __salt__["esxdatacenter.get_details"]()["datacenter"]
    else:
        dc_name = name
    log.info("Running datacenter_configured for datacenter '%s'", dc_name)
    ret = {"name": name, "changes": {}, "result": None, "comment": "Default"}
    comments = []
    si = None
    try:
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        dcs = __salt__["vsphere.list_datacenters_via_proxy"](
            datacenter_names=[dc_name], service_instance=si
        )
        if not dcs:
            if __opts__["test"]:
                comments.append(f"State will create datacenter '{dc_name}'.")
            else:
                log.debug("Creating datacenter '%s'", dc_name)
                __salt__["vsphere.create_datacenter"](dc_name, si)
                comments.append(f"Created datacenter '{dc_name}'.")
            log.info(comments[-1])
            ret["changes"].update({"new": {"name": dc_name}})
        else:
            comments.append(
                f"Datacenter '{dc_name}' already exists. Nothing to be done."
            )
            log.info(comments[-1])
        __salt__["vsphere.disconnect"](si)
        ret["comment"] = "\n".join(comments)
        ret["result"] = None if __opts__["test"] and ret["changes"] else True
        return ret
    except salt.exceptions.CommandExecutionError as exc:
        log.error("Error: %s", exc)
        if si:
            __salt__["vsphere.disconnect"](si)
        ret.update(
            {"result": False if not __opts__["test"] else None, "comment": str(exc)}
        )
        return ret



import logging
import socket
import time

import pytest
from saltfactories.utils import random_string

import salt.cache
import salt.loader
from tests.pytests.functional.cache.helpers import run_common_cache_tests

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


def confirm_consul_is_ready(timeout_at, container):
    sleeptime = 0.1
    consul_port = container.get_host_port_binding(8500, protocol="tcp")
    while time.time() <= timeout_at:
        time.sleep(sleeptime)
        try:
            with socket.create_connection(("localhost", consul_port), timeout=1) as cli:
                cli.send(b"GET /v1/kv/fnord HTTP/1.1\n\n")
                cli.recv(2048)
                break
        except ConnectionResetError as e:
            if e.errno == 104:
                sleeptime += sleeptime
    else:
        return False
    return True


@pytest.fixture(scope="module")
def consul_container(salt_factories):

    container = salt_factories.get_container(
        random_string("consul-server-"),
        image_name="ghcr.io/saltstack/salt-ci-containers/consul:latest",
        container_run_kwargs={"ports": {"8500/tcp": None}},
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    container.container_start_check(confirm_consul_is_ready, container)

    with container.started() as factory:
        yield factory


@pytest.fixture
def cache(minion_opts, consul_container):
    opts = minion_opts.copy()
    opts["cache"] = "consul"
    opts["consul.host"] = "127.0.0.1"
    opts["consul.port"] = consul_container.get_host_port_binding(8500, protocol="tcp")
    # NOTE: If you would like to ensure that alternate suffixes are properly
    # tested, simply change this value and re-run the tests.
    opts["consul.timestamp_suffix"] = ".frobnosticate"
    cache = salt.cache.factory(opts)
    return cache


@pytest.mark.slow_test
def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)



import logging

# pylint: disable=unused-import
from tests.pytests.integration.cluster.conftest import (
    cluster_cache_path,
    cluster_master_1,
    cluster_master_2,
    cluster_master_3,
    cluster_minion_1,
    cluster_pki_path,
    cluster_shared_path,
)

# pylint: enable=unused-import


log = logging.getLogger(__name__)



# this needs to be from! see test_iso_to_timestamp_polyfill
from datetime import datetime

import pytest

import salt.utils.vault.helpers as hlp
from tests.support.mock import patch


@pytest.mark.parametrize(
    "opts_runtype,expected",
    [
        ("master", hlp.SALT_RUNTYPE_MASTER),
        ("master_peer_run", hlp.SALT_RUNTYPE_MASTER_PEER_RUN),
        ("master_impersonating", hlp.SALT_RUNTYPE_MASTER_IMPERSONATING),
        ("minion_local_1", hlp.SALT_RUNTYPE_MINION_LOCAL),
        ("minion_local_2", hlp.SALT_RUNTYPE_MINION_LOCAL),
        ("minion_local_3", hlp.SALT_RUNTYPE_MINION_LOCAL),
        ("minion_remote", hlp.SALT_RUNTYPE_MINION_REMOTE),
    ],
    indirect=["opts_runtype"],
)
def test_get_salt_run_type(opts_runtype, expected):
    """
    Ensure run types are detected as expected
    """
    assert hlp._get_salt_run_type(opts_runtype) == expected


@pytest.mark.parametrize(
    "pattern,expected",
    [
        ("no-tokens-to-replace", ["no-tokens-to-replace"]),
        ("single-dict:{minion}", ["single-dict:{minion}"]),
        ("single-list:{grains[roles]}", ["single-list:web", "single-list:database"]),
        (
            "multiple-lists:{grains[roles]}+{grains[aux]}",
            [
                "multiple-lists:web+foo",
                "multiple-lists:web+bar",
                "multiple-lists:database+foo",
                "multiple-lists:database+bar",
            ],
        ),
        (
            "single-list-with-dicts:{grains[id]}+{grains[roles]}+{grains[id]}",
            [
                "single-list-with-dicts:{grains[id]}+web+{grains[id]}",
                "single-list-with-dicts:{grains[id]}+database+{grains[id]}",
            ],
        ),
        (
            "deeply-nested-list:{grains[deep][foo][bar][baz]}",
            [
                "deeply-nested-list:hello",
                "deeply-nested-list:world",
            ],
        ),
    ],
)
def test_expand_pattern_lists(pattern, expected):
    """
    Ensure expand_pattern_lists works as intended:
    - Expand list-valued patterns
    - Do not change non-list-valued tokens
    """
    pattern_vars = {
        "id": "test-minion",
        "roles": ["web", "database"],
        "aux": ["foo", "bar"],
        "deep": {"foo": {"bar": {"baz": ["hello", "world"]}}},
    }

    mappings = {"minion": "test-minion", "grains": pattern_vars}
    output = hlp.expand_pattern_lists(pattern, **mappings)
    assert output == expected


@pytest.mark.parametrize(
    "inpt,expected",
    [
        (60.0, 60.0),
        (60, 60.0),
        ("60", 60.0),
        ("60s", 60.0),
        ("2m", 120.0),
        ("1h", 3600.0),
        ("1d", 86400.0),
        ("1.5s", 1.5),
        ("1.5m", 90.0),
        ("1.5h", 5400.0),
        ("7.5d", 648000.0),
    ],
)
def test_timestring_map(inpt, expected):
    assert hlp.timestring_map(inpt) == expected


@pytest.mark.parametrize(
    "creation_time,expected",
    [
        ("2022-08-22T17:16:21-09:30", 1661222781),
        ("2022-08-22T17:16:21-01:00", 1661192181),
        ("2022-08-22T17:16:21+00:00", 1661188581),
        ("2022-08-22T17:16:21Z", 1661188581),
        ("2022-08-22T17:16:21+02:00", 1661181381),
        ("2022-08-22T17:16:21+12:30", 1661143581),
    ],
)
def test_iso_to_timestamp_polyfill(creation_time, expected):
    with patch("salt.utils.vault.helpers.datetime.datetime") as d:
        d.fromisoformat.side_effect = AttributeError
        # needs from datetime import datetime, otherwise results
        # in infinite recursion

        # pylint: disable=unnecessary-lambda
        d.side_effect = lambda *args: datetime(*args)
        res = hlp.iso_to_timestamp(creation_time)
        assert res == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.process as process
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {process: {}}


def test_absent():
    """
    Test to ensures that the named command is not running.
    """
    name = "apache2"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    mock = MagicMock(return_value="")
    with patch.dict(process.__salt__, {"ps.pgrep": mock, "ps.pkill": mock}):
        with patch.dict(process.__opts__, {"test": True}):
            comt = "No matching processes running"
            ret.update({"comment": comt})
            assert process.absent(name) == ret

        with patch.dict(process.__opts__, {"test": False}):
            ret.update({"result": True})
            assert process.absent(name) == ret



import logging
import random
import string

import pytest

import salt.config
import salt.loader
import salt.states.boto_elasticsearch_domain as boto_elasticsearch_domain
from tests.support.mock import MagicMock, patch

boto = pytest.importorskip("boto")
boto3 = pytest.importorskip("boto3", "1.2.1")
botocore = pytest.importorskip("botocore", "1.4.41")

log = logging.getLogger(__name__)


class GlobalConfig:
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = botocore.exceptions.ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
                "Message": "Test-defined error",
            }
        },
        "msg",
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    domain_ret = dict(
        DomainName="testdomain",
        ElasticsearchClusterConfig={},
        EBSOptions={},
        AccessPolicies={},
        SnapshotOptions={},
        AdvancedOptions={},
        ElasticsearchVersion="1.5",
    )


@pytest.fixture
def session_instance():
    GlobalConfig.conn_parameters["key"] = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
    )
    with patch("boto3.session.Session") as patched_session:
        yield patched_session()


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts["grains"] = salt.loader.grains(minion_opts)
    ctx = {}
    utils = salt.loader.utils(
        minion_opts,
        whitelist=["boto3", "args", "systemd", "path", "platform"],
        context=ctx,
    )
    serializers = salt.loader.serializers(minion_opts)
    funcs = funcs = salt.loader.minion_mods(
        minion_opts, context=ctx, utils=utils, whitelist=["boto_elasticsearch_domain"]
    )
    salt_states = salt.loader.states(
        opts=minion_opts,
        functions=funcs,
        utils=utils,
        whitelist=["boto_elasticsearch_domain"],
        serializers=serializers,
    )
    return {
        boto_elasticsearch_domain: {
            "__opts__": minion_opts,
            "__salt__": funcs,
            "__utils__": utils,
            "__states__": salt_states,
            "__serializers__": serializers,
        }
    }


def test_present_when_domain_does_not_exist(session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.side_effect = GlobalConfig.not_found_error
    conn.describe_elasticsearch_domain_config.return_value = {
        "DomainConfig": GlobalConfig.domain_ret
    }
    conn.create_elasticsearch_domain.return_value = {
        "DomainStatus": GlobalConfig.domain_ret
    }
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.present"](
        "domain present", **GlobalConfig.domain_ret
    )

    assert result["result"]
    assert result["changes"]["new"]["domain"]["ElasticsearchClusterConfig"] is None


def test_present_when_domain_exists(session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.return_value = {
        "DomainStatus": GlobalConfig.domain_ret
    }
    cfg = {}
    for k, v in GlobalConfig.domain_ret.items():
        cfg[k] = {"Options": v}
    cfg["AccessPolicies"] = {"Options": '{"a": "b"}'}
    conn.describe_elasticsearch_domain_config.return_value = {"DomainConfig": cfg}
    conn.update_elasticsearch_domain_config.return_value = {"DomainConfig": cfg}
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.present"](
        "domain present", **GlobalConfig.domain_ret
    )
    assert result["result"]
    assert result["changes"] == {
        "new": {"AccessPolicies": {}},
        "old": {"AccessPolicies": {"a": "b"}},
    }


def test_present_with_failure(session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.side_effect = GlobalConfig.not_found_error
    conn.describe_elasticsearch_domain_config.return_value = {
        "DomainConfig": GlobalConfig.domain_ret
    }
    conn.create_elasticsearch_domain.side_effect = botocore.exceptions.ClientError(
        GlobalConfig.error_content, "create_domain"
    )
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.present"](
        "domain present", **GlobalConfig.domain_ret
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_absent_when_domain_does_not_exist(session_instance):
    """
    Tests absent on a domain that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.side_effect = GlobalConfig.not_found_error
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.absent"](
        "test", "mydomain"
    )
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_domain_exists(session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.return_value = {
        "DomainStatus": GlobalConfig.domain_ret
    }
    conn.describe_elasticsearch_domain_config.return_value = {
        "DomainConfig": GlobalConfig.domain_ret
    }
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.absent"](
        "test", GlobalConfig.domain_ret["DomainName"]
    )
    assert result["result"]
    assert result["changes"]["new"]["domain"] is None


def test_absent_with_failure(session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn
    conn.describe_elasticsearch_domain.return_value = {
        "DomainStatus": GlobalConfig.domain_ret
    }
    conn.describe_elasticsearch_domain_config.return_value = {
        "DomainConfig": GlobalConfig.domain_ret
    }
    conn.delete_elasticsearch_domain.side_effect = botocore.exceptions.ClientError(
        GlobalConfig.error_content, "delete_domain"
    )
    result = boto_elasticsearch_domain.__states__["boto_elasticsearch_domain.absent"](
        "test", GlobalConfig.domain_ret["DomainName"]
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.dpkg
"""

import logging
import os

import pytest

import salt.modules.dpkg_lowpkg as dpkg
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {dpkg: {}}


def setUp(self):
    dpkg_lowpkg_logger = logging.getLogger("salt.modules.dpkg_lowpkg")
    self.level = dpkg_lowpkg_logger.level
    dpkg_lowpkg_logger.setLevel(logging.FATAL)


def tearDown(self):
    logging.getLogger("salt.modules.dpkg_lowpkg").setLevel(self.level)


def dpkg_L_side_effect(cmd, **kwargs):
    assert cmd[:2] == ["dpkg", "-L"]
    package = cmd[2]
    return dpkg_l_output[package]


dpkg_error_msg = """dpkg-query: package 'httpd' is not installed
Use dpkg --contents (= dpkg-deb --contents) to list archive files contents.
"""


dpkg_l_output = {
    "hostname": """\
/.
/bin
/bin/hostname
/usr
/usr/share
/usr/share/doc
/usr/share/doc/hostname
/usr/share/doc/hostname/changelog.gz
/usr/share/doc/hostname/copyright
/usr/share/man
/usr/share/man/man1
/usr/share/man/man1/hostname.1.gz
/bin/dnsdomainname
/bin/domainname
/bin/nisdomainname
/bin/ypdomainname
/usr/share/man/man1/dnsdomainname.1.gz
/usr/share/man/man1/domainname.1.gz
/usr/share/man/man1/nisdomainname.1.gz
/usr/share/man/man1/ypdomainname.1.gz
"""
}


# 'unpurge' function tests: 2


def test_unpurge():
    """
    Test if it change package selection for each package
    specified to 'install'
    """
    mock = MagicMock(return_value=[])
    with patch.dict(dpkg.__salt__, {"pkg.list_pkgs": mock, "cmd.run": mock}):
        assert dpkg.unpurge("curl") == {}


def test_unpurge_empty_package():
    """
    Test if it change package selection for each package
    specified to 'install'
    """
    assert dpkg.unpurge() == {}


# 'list_pkgs' function tests: 1


def test_list_pkgs():
    """
    Test if it lists the packages currently installed
    """
    mock = MagicMock(
        return_value={
            "retcode": 0,
            "stderr": "",
            "stdout": "installed\thostname\t3.21",
        }
    )
    with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
        assert dpkg.list_pkgs("hostname") == {"hostname": "3.21"}

    mock = MagicMock(
        return_value={
            "retcode": 1,
            "stderr": "dpkg-query: no packages found matching httpd",
            "stdout": "",
        }
    )
    with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
        assert (
            dpkg.list_pkgs("httpd")
            == "Error:  dpkg-query: no packages found matching httpd"
        )


# 'file_list' function tests: 1


def test_file_list():
    """
    Test if it lists the files that belong to a package.
    """
    dpkg_query_mock = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": "installed\thostname"}
    )
    dpkg_L_mock = MagicMock(side_effect=dpkg_L_side_effect)
    with patch.dict(
        dpkg.__salt__, {"cmd.run_all": dpkg_query_mock, "cmd.run": dpkg_L_mock}
    ):
        assert dpkg.file_list("hostname") == {
            "errors": [],
            "files": [
                "/.",
                "/bin",
                "/bin/dnsdomainname",
                "/bin/domainname",
                "/bin/hostname",
                "/bin/nisdomainname",
                "/bin/ypdomainname",
                "/usr",
                "/usr/share",
                "/usr/share/doc",
                "/usr/share/doc/hostname",
                "/usr/share/doc/hostname/changelog.gz",
                "/usr/share/doc/hostname/copyright",
                "/usr/share/man",
                "/usr/share/man/man1",
                "/usr/share/man/man1/dnsdomainname.1.gz",
                "/usr/share/man/man1/domainname.1.gz",
                "/usr/share/man/man1/hostname.1.gz",
                "/usr/share/man/man1/nisdomainname.1.gz",
                "/usr/share/man/man1/ypdomainname.1.gz",
            ],
        }

    mock = MagicMock(
        return_value={"retcode": 1, "stderr": dpkg_error_msg, "stdout": ""}
    )
    with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
        assert dpkg.file_list("httpd") == "Error:  " + dpkg_error_msg


# 'file_dict' function tests: 1


def test_file_dict():
    """
    Test if it lists the files that belong to a package, grouped by package
    """
    dpkg_query_mock = MagicMock(
        return_value={"retcode": 0, "stderr": "", "stdout": "installed\thostname"}
    )
    dpkg_L_mock = MagicMock(side_effect=dpkg_L_side_effect)
    with patch.dict(
        dpkg.__salt__, {"cmd.run_all": dpkg_query_mock, "cmd.run": dpkg_L_mock}
    ):
        expected = {
            "errors": [],
            "packages": {
                "hostname": [
                    "/.",
                    "/bin",
                    "/bin/hostname",
                    "/usr",
                    "/usr/share",
                    "/usr/share/doc",
                    "/usr/share/doc/hostname",
                    "/usr/share/doc/hostname/changelog.gz",
                    "/usr/share/doc/hostname/copyright",
                    "/usr/share/man",
                    "/usr/share/man/man1",
                    "/usr/share/man/man1/hostname.1.gz",
                    "/bin/dnsdomainname",
                    "/bin/domainname",
                    "/bin/nisdomainname",
                    "/bin/ypdomainname",
                    "/usr/share/man/man1/dnsdomainname.1.gz",
                    "/usr/share/man/man1/domainname.1.gz",
                    "/usr/share/man/man1/nisdomainname.1.gz",
                    "/usr/share/man/man1/ypdomainname.1.gz",
                ]
            },
        }
        assert dpkg.file_dict("hostname") == expected

    mock = MagicMock(
        return_value={"retcode": 1, "stderr": dpkg_error_msg, "stdout": ""}
    )
    with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}):
        assert dpkg.file_dict("httpd") == "Error:  " + dpkg_error_msg


def test_bin_pkg_info_spaces():
    """
    Test the bin_pkg_info function
    """
    file_proto_mock = MagicMock(return_value=True)
    with patch.dict(dpkg.__salt__, {"config.valid_fileproto": file_proto_mock}):
        cache_mock = MagicMock(return_value="/path/to/some/package.deb")
        with patch.dict(dpkg.__salt__, {"cp.cache_file": cache_mock}):
            dpkg_info_mock = MagicMock(
                return_value={
                    "retcode": 0,
                    "stderr": "",
                    "stdout": (
                        " new Debian package, version 2.0\n"
                        " size 123456 bytes: control archive: 4029  bytes.\n"
                        " Package          : package_name\n"
                        " Version          : 1.0\n"
                        " Section          : section_name\n"
                        " Priority         : priority\n"
                        " Architecture     : all\n"
                        " Description      : some package\n"
                    ),
                }
            )
            with patch.dict(dpkg.__salt__, {"cmd.run_all": dpkg_info_mock}):
                assert dpkg.bin_pkg_info("package.deb")["name"] == "package_name"


def test_bin_pkg_info_no_spaces():
    """
    Test the bin_pkg_info function
    """
    file_proto_mock = MagicMock(return_value=True)
    with patch.dict(dpkg.__salt__, {"config.valid_fileproto": file_proto_mock}):
        cache_mock = MagicMock(return_value="/path/to/some/package.deb")
        with patch.dict(dpkg.__salt__, {"cp.cache_file": cache_mock}):
            dpkg_info_mock = MagicMock(
                return_value={
                    "retcode": 0,
                    "stderr": "",
                    "stdout": (
                        " new Debian package, version 2.0\n"
                        " size 123456 bytes: control archive: 4029  bytes.\n"
                        " Package: package_name\n"
                        " Version: 1.0\n"
                        " Section: section_name\n"
                        " Priority: priority\n"
                        " Architecture: all\n"
                        " Description: some package\n"
                    ),
                }
            )
            with patch.dict(dpkg.__salt__, {"cmd.run_all": dpkg_info_mock}):
                assert dpkg.bin_pkg_info("package.deb")["name"] == "package_name"


def test_info():
    """
    Test package info
    """
    mock = MagicMock(
        return_value={
            "retcode": 0,
            "stderr": "",
            "stdout": os.linesep.join(
                [
                    "package:bash",
                    "revision:",
                    "architecture:amd64",
                    "maintainer:Ubuntu Developers"
                    " <ubuntu-devel-discuss@lists.ubuntu.com>",
                    "summary:",
                    "source:bash",
                    "version:4.4.18-2ubuntu1",
                    "section:shells",
                    "installed_size:1588",
                    "size:",
                    "MD5:",
                    "SHA1:",
                    "SHA256:",
                    "origin:",
                    "homepage:http://tiswww.case.edu/php/chet/bash/bashtop.html",
                    "status:ii ",
                    "description:GNU Bourne Again SHell",
                    " Bash is an sh-compatible command language interpreter that"
                    " executes",
                    " commands read from the standard input or from a file.  Bash"
                    " also",
                    " incorporates useful features from the Korn and C shells (ksh"
                    " and csh).",
                    " .",
                    " Bash is ultimately intended to be a conformant implementation"
                    " of the",
                    " IEEE POSIX Shell and Tools specification (IEEE Working Group"
                    " 1003.2).",
                    " .",
                    " The Programmable Completion Code, by Ian Macdonald, is now"
                    " found in",
                    " the bash-completion package.",
                    "",
                    "*/~^\\*",  # pylint: disable=W1401
                ]
            ),
        }
    )

    with patch.dict(dpkg.__salt__, {"cmd.run_all": mock}), patch.dict(
        dpkg.__grains__, {"os": "Ubuntu", "osrelease_info": (18, 4)}
    ), patch("salt.utils.path.which", MagicMock(return_value=False)), patch(
        "os.path.exists", MagicMock(return_value=False)
    ), patch(
        "os.path.getmtime", MagicMock(return_value=1560199259.0)
    ):
        assert dpkg.info("bash") == {
            "bash": {
                "architecture": "amd64",
                "description": os.linesep.join(
                    [
                        "GNU Bourne Again SHell",
                        " Bash is an sh-compatible command language interpreter"
                        " that executes",
                        " commands read from the standard input or from a file."
                        "  Bash also",
                        " incorporates useful features from the Korn and C"
                        " shells (ksh and csh).",
                        " .",
                        " Bash is ultimately intended to be a conformant"
                        " implementation of the",
                        " IEEE POSIX Shell and Tools specification (IEEE"
                        " Working Group 1003.2).",
                        " .",
                        " The Programmable Completion Code, by Ian Macdonald,"
                        " is now found in",
                        " the bash-completion package." + os.linesep,
                    ]
                ),
                "homepage": "http://tiswww.case.edu/php/chet/bash/bashtop.html",
                "maintainer": (
                    "Ubuntu Developers <ubuntu-devel-discuss@lists.ubuntu.com>"
                ),
                "package": "bash",
                "section": "shells",
                "source": "bash",
                "status": "ii",
                "version": "4.4.18-2ubuntu1",
            }
        }


def test_get_pkg_license():
    """
    Test _get_pkg_license for ignore errors on reading license from copyright files
    """
    license_read_mock = mock_open(read_data="")
    with patch.object(os.path, "exists", MagicMock(return_value=True)), patch(
        "salt.utils.files.fopen", license_read_mock
    ):
        dpkg._get_pkg_license("bash")

        assert license_read_mock.calls[0].args[0] == "/usr/share/doc/bash/copyright"
        assert license_read_mock.calls[0].kwargs["errors"] == "ignore"



"""
Management of LVS (Linux Virtual Server) Real Server
====================================================
"""


def __virtual__():
    """
    Only load if the lvs module is available in __salt__
    """
    if "lvs.get_rules" in __salt__:
        return "lvs_server"
    return (False, "lvs module could not be loaded")


def present(
    name,
    protocol=None,
    service_address=None,
    server_address=None,
    packet_forward_method="dr",
    weight=1,
):
    """
    Ensure that the named service is present.

    name
        The LVS server name

    protocol
        The service protocol

    service_address
        The LVS service address

    server_address
        The real server address.

    packet_forward_method
        The LVS packet forwarding method(``dr`` for direct routing, ``tunnel`` for tunneling, ``nat`` for network access translation).

    weight
        The capacity  of a server relative to the others in the pool.


    .. code-block:: yaml

        lvsrs:
          lvs_server.present:
            - protocol: tcp
            - service_address: 1.1.1.1:80
            - server_address: 192.168.0.11:8080
            - packet_forward_method: dr
            - weight: 10
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # check server
    server_check = __salt__["lvs.check_server"](
        protocol=protocol,
        service_address=service_address,
        server_address=server_address,
    )
    if server_check is True:
        server_rule_check = __salt__["lvs.check_server"](
            protocol=protocol,
            service_address=service_address,
            server_address=server_address,
            packet_forward_method=packet_forward_method,
            weight=weight,
        )
        if server_rule_check is True:
            ret["comment"] = "LVS Server {} in service {}({}) is present".format(
                name, service_address, protocol
            )
            return ret
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = (
                    "LVS Server {} in service {}({}) is present but some options should"
                    " update".format(name, service_address, protocol)
                )
                return ret
            else:
                server_edit = __salt__["lvs.edit_server"](
                    protocol=protocol,
                    service_address=service_address,
                    server_address=server_address,
                    packet_forward_method=packet_forward_method,
                    weight=weight,
                )
                if server_edit is True:
                    ret["comment"] = (
                        "LVS Server {} in service {}({}) has been updated".format(
                            name, service_address, protocol
                        )
                    )
                    ret["changes"][name] = "Update"
                    return ret
                else:
                    ret["result"] = False
                    ret["comment"] = (
                        "LVS Server {} in service {}({}) update failed({})".format(
                            name, service_address, protocol, server_edit
                        )
                    )
                    return ret
    else:
        if __opts__["test"]:
            ret["comment"] = (
                "LVS Server {} in service {}({}) is not present and needs to be created".format(
                    name, service_address, protocol
                )
            )
            ret["result"] = None
            return ret
        else:
            server_add = __salt__["lvs.add_server"](
                protocol=protocol,
                service_address=service_address,
                server_address=server_address,
                packet_forward_method=packet_forward_method,
                weight=weight,
            )
            if server_add is True:
                ret["comment"] = (
                    "LVS Server {} in service {}({}) has been created".format(
                        name, service_address, protocol
                    )
                )
                ret["changes"][name] = "Present"
                return ret
            else:
                ret["comment"] = (
                    "LVS Service {} in service {}({}) create failed({})".format(
                        name, service_address, protocol, server_add
                    )
                )
                ret["result"] = False
                return ret


def absent(name, protocol=None, service_address=None, server_address=None):
    """
    Ensure the LVS Real Server in specified service is absent.

    name
        The name of the LVS server.

    protocol
        The service protocol(only support ``tcp``, ``udp`` and ``fwmark`` service).

    service_address
        The LVS service address.

    server_address
        The LVS real server address.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # check if server exists and remove it
    server_check = __salt__["lvs.check_server"](
        protocol=protocol,
        service_address=service_address,
        server_address=server_address,
    )
    if server_check is True:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                "LVS Server {} in service {}({}) is present and needs to be removed".format(
                    name, service_address, protocol
                )
            )
            return ret
        server_delete = __salt__["lvs.delete_server"](
            protocol=protocol,
            service_address=service_address,
            server_address=server_address,
        )
        if server_delete is True:
            ret["comment"] = "LVS Server {} in service {}({}) has been removed".format(
                name, service_address, protocol
            )
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["comment"] = (
                "LVS Server {} in service {}({}) removed failed({})".format(
                    name, service_address, protocol, server_delete
                )
            )
            ret["result"] = False
            return ret
    else:
        ret["comment"] = (
            "LVS Server {} in service {}({}) is not present, so it cannot be removed".format(
                name, service_address, protocol
            )
        )

    return ret



"""
Tests for the salt-run command
"""

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


def test_master(salt_run_cli, salt_minion):
    """
    jobs.master
    """
    ret = salt_run_cli.run("jobs.master")
    assert ret.data == []
    assert ret.stdout.strip() == "[]"


def test_active(salt_run_cli, salt_minion):
    """
    jobs.active
    """
    ret = salt_run_cli.run("jobs.active")
    assert ret.data == {}
    assert ret.stdout.strip() == "{}"


def test_lookup_jid(salt_run_cli, salt_minion):
    """
    jobs.lookup_jid
    """
    ret = salt_run_cli.run("jobs.lookup_jid", "23974239742394")
    assert ret.data == {}
    assert ret.stdout.strip() == "{}"


def test_lookup_jid_invalid(salt_run_cli, salt_minion):
    """
    jobs.lookup_jid
    """
    ret = salt_run_cli.run("jobs.lookup_jid")
    expected = "Passed invalid arguments:"
    assert expected in ret.stdout


def test_list_jobs(salt_run_cli, salt_minion, salt_cli):
    """
    jobs.list_jobs
    """
    salt_cli.run("test.echo", "test_list_jobs", minion_tgt=salt_minion.id)
    ret = salt_run_cli.run("jobs.list_jobs")
    assert isinstance(ret.data, dict)
    for job in ret.data.values():
        if job["Function"] != "test.echo":
            continue
        if job["Arguments"] != ["test_list_jobs"]:
            continue
        # We found our job in the list, we're good with the test
        break
    else:
        pytest.fail("Did not find our job from the jobs.list_jobs call")


def test_target_info(salt_run_cli, salt_minion, salt_cli):
    """
    This is a test case for issue #48734

    PR #43454 fixed an issue where "jobs.lookup_jid" was not working
    correctly with external job caches. However, this fix for external
    job caches broke some inner workings of job storage when using the
    local_cache.

    We need to preserve the previous behavior for the local_cache, but
    keep the new behavior for other external job caches.

    If "savefstr" is called in the local cache, the target data does not
    get written to the local_cache, and the target-type gets listed as a
    "list" type instead of "glob".

    This is a regression test for fixing the local_cache behavior.
    """
    salt_cli.run("test.echo", "target_info_test", minion_tgt=salt_minion.id)
    ret = salt_run_cli.run("jobs.list_jobs")
    for item in ret.data.values():
        if (
            item["Function"] == "test.echo"
            and item["Arguments"][0] == "target_info_test"
        ):
            job_ret = item
    tgt = job_ret["Target"]
    tgt_type = job_ret["Target-type"]

    assert tgt != "unknown-target"
    assert tgt == salt_minion.id
    assert tgt_type == "glob"


def test_jobs_runner(salt_run_cli, salt_minion):
    """
    Test when running a runner job and then
    running jobs_list to ensure the job was saved
    properly in the cache
    """
    salt_run_cli.run("test.arg", "arg1", kwarg1="kwarg1")
    ret = salt_run_cli.run("jobs.list_jobs")
    jid = None
    for key, item in ret.data.items():
        if item["Function"] == "runner.test.arg":
            jid = key

    get_job = salt_run_cli.run("jobs.list_job", jid)
    assert not get_job.data.get("Error")
    assert get_job.data["jid"] == jid


def test_target_info_salt_call(salt_run_cli, salt_minion, salt_call_cli):
    """
    Check the job infor for a call initiated
    with salt call
    """
    test = salt_call_cli.run("test.echo", "target_info_test", minion_tgt=salt_minion.id)
    ret = salt_run_cli.run("jobs.list_jobs")
    for item in ret.data.values():
        if (
            item["Function"] == "test.echo"
            and item["Arguments"][0] == "target_info_test"
        ):
            job_ret = item
    tgt = job_ret["Target"]
    tgt_type = job_ret["Target-type"]

    assert tgt != "unknown-target"
    assert tgt == salt_minion.id
    assert tgt_type == "glob"



import pytest

import salt.utils.vault as vault
import salt.utils.vault.auth as vauth
import salt.utils.vault.cache as vcache
import salt.utils.vault.client as vclient
import salt.utils.vault.leases as vleases
from tests.support.mock import Mock, patch


@pytest.fixture
def token(token_auth):
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def token_invalid(token_auth):
    token_auth["auth"]["num_uses"] = 1
    token_auth["auth"]["use_count"] = 1
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def token_unrenewable(token_auth):
    token_auth["auth"]["renewable"] = False
    return vleases.VaultToken(**token_auth["auth"])


@pytest.fixture
def secret_id(secret_id_response):
    return vleases.VaultSecretId(**secret_id_response["data"])


@pytest.fixture
def secret_id_invalid(secret_id_response):
    secret_id_response["data"]["secret_id_num_uses"] = 1
    secret_id_response["data"]["use_count"] = 1
    return vleases.VaultSecretId(**secret_id_response["data"])


@pytest.fixture(params=["secret_id"])
def approle(request):
    secret_id = request.param
    if secret_id is not None:
        secret_id = request.getfixturevalue(secret_id)
    return vauth.VaultAppRole("test-role-id", secret_id)


@pytest.fixture
def approle_invalid(secret_id_invalid):
    return vauth.VaultAppRole("test-role-id", secret_id_invalid)


@pytest.fixture
def token_store(token):
    store = Mock(spec=vauth.VaultTokenAuth)
    store.is_valid.return_value = True
    store.get_token.return_value = token
    return store


@pytest.fixture
def token_store_empty(token_store):
    token_store.is_valid.return_value = False
    token_store.get_token.side_effect = vault.VaultAuthExpired
    return token_store


@pytest.fixture
def token_store_empty_first(token_store, token):
    token_store.is_valid.side_effect = (False, True)
    token_store.get_token.side_effect = (token, vault.VaultException)
    return token_store


@pytest.fixture
def uncached():
    cache = Mock(spec=vcache.VaultAuthCache)
    cache.exists.return_value = False
    cache.get.return_value = None
    return cache


@pytest.fixture
def cached_token(uncached, token):
    uncached.exists.return_value = True
    uncached.get.return_value = token
    return uncached


@pytest.fixture
def client(token_auth):
    token_auth["auth"]["client_token"] = "new-test-token"
    client = Mock(spec=vclient.VaultClient)
    client.post.return_value = token_auth
    return client


def test_token_auth_uninitialized(uncached):
    """
    Test that an exception is raised when a token is requested
    and the authentication container was not passed a valid token.
    """
    auth = vauth.VaultTokenAuth(cache=uncached)
    uncached.get.assert_called_once()
    assert auth.is_valid() is False
    assert auth.is_renewable() is False
    auth.used()
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


def test_token_auth_cached(cached_token, token):
    """
    Test that tokens are read from cache.
    """
    auth = vauth.VaultTokenAuth(cache=cached_token)
    assert auth.is_valid()
    assert auth.get_token() == token


def test_token_auth_invalid_token(invalid_token):
    """
    Test that an exception is raised when a token is requested
    and the container's token is invalid.
    """
    auth = vauth.VaultTokenAuth(token=invalid_token)
    assert auth.is_valid() is False
    assert auth.is_renewable() is False
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


def test_token_auth_unrenewable_token(token_unrenewable):
    """
    Test that it is reported correctly by the container
    when a token is not renewable.
    """
    auth = vauth.VaultTokenAuth(token=token_unrenewable)
    assert auth.is_valid() is True
    assert auth.is_renewable() is False
    assert auth.get_token() == token_unrenewable


@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_token_auth_used_num_uses(uncached, token, num_uses):
    """
    Ensure that cache writes for use count are only done when
    num_uses is not 0 (= unlimited).
    Single-use tokens still require cache writes for updating
    ``uses``. The cache cannot be flushed here since
    exceptions might be used to indicate the token expiry
    to factory methods.
    """
    token = token.with_renewed(num_uses=num_uses)
    auth = vauth.VaultTokenAuth(cache=uncached, token=token)
    auth.used()
    if num_uses > 0:
        uncached.store.assert_called_once_with(token)
    else:
        uncached.store.assert_not_called()


@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_token_auth_update_token(uncached, token, num_uses):
    """
    Ensure that partial updates to the token in use are possible
    and that the cache writes are independent from num_uses.
    Also ensure the token is treated as immutable
    """
    auth = vauth.VaultTokenAuth(cache=uncached, token=token)
    old_token = token
    old_token_ttl = old_token.duration
    auth.update_token({"num_uses": num_uses, "ttl": 8483})
    updated_token = token.with_renewed(num_uses=num_uses, ttl=8483)
    assert auth.token == updated_token
    assert old_token.duration == old_token_ttl
    uncached.store.assert_called_once_with(updated_token)


def test_token_auth_replace_token(uncached, token):
    """
    Ensure completely replacing the token is possible and
    results in a cache write. This is important when an
    InvalidVaultToken has to be replaced with a VaultToken,
    eg by a different authentication method.
    """
    auth = vauth.VaultTokenAuth(cache=uncached)
    assert isinstance(auth.token, vauth.InvalidVaultToken)
    auth.replace_token(token)
    assert isinstance(auth.token, vleases.VaultToken)
    assert auth.token == token
    uncached.store.assert_called_once_with(token)


@pytest.mark.parametrize("token", [False, True])
@pytest.mark.parametrize("approle", [False, True])
def test_approle_auth_is_valid(token, approle):
    """
    Test that is_valid reports true when either the token
    or the secret ID is valid
    """
    token = Mock(spec=vleases.VaultToken)
    token.is_valid.return_value = token
    approle = Mock(spec=vleases.VaultSecretId)
    approle.is_valid.return_value = approle
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token)
    assert auth.is_valid() is (token or approle)


def test_approle_auth_get_token_store_available(token_store, approle, token):
    """
    Ensure no login attempt is made when a cached token is available
    """
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token_store)
    with patch("salt.utils.vault.auth.VaultAppRoleAuth._login") as login:
        res = auth.get_token()
        login.assert_not_called()
        assert res == token


def test_approle_auth_get_token_store_empty(token_store_empty, approle, token):
    """
    Ensure a token is returned if no cached token is available
    """
    auth = vauth.VaultAppRoleAuth(approle, None, token_store=token_store_empty)
    with patch("salt.utils.vault.auth.VaultAppRoleAuth._login") as login:
        login.return_value = token
        res = auth.get_token()
        login.assert_called_once()
        assert res == token


def test_approle_auth_get_token_invalid(token_store_empty, approle_invalid):
    """
    Ensure VaultAuthExpired is raised if a token request was made, but
    cannot be fulfilled
    """
    auth = vauth.VaultAppRoleAuth(approle_invalid, None, token_store=token_store_empty)
    with pytest.raises(vault.VaultAuthExpired):
        auth.get_token()


@pytest.mark.parametrize("mount", ["approle", "salt_minions"])
@pytest.mark.parametrize("approle", ["secret_id", None], indirect=True)
def test_approle_auth_get_token_login(
    approle, mount, client, token_store_empty_first, token
):
    """
    Ensure that login with secret-id returns a token that is passed to the
    token store/cache as well
    """
    auth = vauth.VaultAppRoleAuth(
        approle, client, mount=mount, token_store=token_store_empty_first
    )
    res = auth.get_token()
    assert res == token
    args, kwargs = client.post.call_args
    endpoint = args[0]
    payload = kwargs.get("payload", {})
    assert endpoint == f"auth/{mount}/login"
    assert "role_id" in payload
    if approle.secret_id is not None:
        assert "secret_id" in payload
    token_store_empty_first.replace_token.assert_called_once_with(res)


@pytest.mark.parametrize("num_uses", [0, 1, 10])
def test_approle_auth_used_num_uses(
    token_store_empty_first, approle, client, uncached, num_uses, token
):
    """
    Ensure that cache writes for use count are only done when
    num_uses is not 0 (= unlimited)
    """
    approle.secret_id = approle.secret_id.with_renewed(num_uses=num_uses)
    auth = vauth.VaultAppRoleAuth(
        approle, client, cache=uncached, token_store=token_store_empty_first
    )
    res = auth.get_token()
    assert res == token
    if num_uses > 1:
        uncached.store.assert_called_once_with(approle.secret_id)
    elif num_uses:
        uncached.store.assert_not_called()
        uncached.flush.assert_called_once()
    else:
        uncached.store.assert_not_called()


def test_approle_auth_used_locally_configured(
    token_store_empty_first, approle, client, uncached, token
):
    """
    Ensure that locally configured secret IDs are not cached.
    """
    approle.secret_id = vault.LocalVaultSecretId(**approle.secret_id.to_dict())
    auth = vauth.VaultAppRoleAuth(
        approle, client, cache=uncached, token_store=token_store_empty_first
    )
    res = auth.get_token()
    assert res == token
    uncached.store.assert_not_called()


def test_approle_allows_no_secret_id():
    """
    Ensure AppRole containers are still valid if no
    secret ID has been set (bind_secret_id can be set to False!)
    """
    role = vauth.VaultAppRole("test-role-id")
    assert role.is_valid()



"""
tests.pytests.unit.test_crypt
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for salt's crypt module
"""

import uuid

import pytest

import salt.crypt
import salt.master
import salt.payload
import salt.utils.files
from tests.conftest import FIPS_TESTRUN
from tests.support.helpers import dedent

from . import PRIV_KEY, PRIV_KEY2, PUB_KEY, PUB_KEY2


def test_get_rsa_pub_key_bad_key(tmp_path):
    """
    get_rsa_pub_key raises InvalidKeyError when encoutering a bad key
    """
    key_path = str(tmp_path / "key")
    with salt.utils.files.fopen(key_path, "w") as fp:
        fp.write("")
    with pytest.raises(salt.crypt.InvalidKeyError):
        salt.crypt.get_rsa_pub_key(key_path)


def test_cryptical_dumps_no_nonce():
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data)

    # Validate message structure
    assert isinstance(ret, bytes)
    une = master_crypt.decrypt(ret)
    une.startswith(master_crypt.PICKLE_PAD)
    assert salt.payload.loads(une[len(master_crypt.PICKLE_PAD) :]) == data

    # Validate load back to orig data
    assert master_crypt.loads(ret) == data


def test_cryptical_dumps_valid_nonce():
    nonce = uuid.uuid4().hex
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data, nonce=nonce)

    assert isinstance(ret, bytes)
    une = master_crypt.decrypt(ret)
    une.startswith(master_crypt.PICKLE_PAD)
    nonce_and_data = une[len(master_crypt.PICKLE_PAD) :]
    assert nonce_and_data.startswith(nonce.encode())
    assert salt.payload.loads(nonce_and_data[len(nonce) :]) == data

    assert master_crypt.loads(ret, nonce=nonce) == data


def test_cryptical_dumps_invalid_nonce():
    nonce = uuid.uuid4().hex
    master_crypt = salt.crypt.Crypticle({}, salt.crypt.Crypticle.generate_key_string())
    data = {"foo": "bar"}
    ret = master_crypt.dumps(data, nonce=nonce)
    assert isinstance(ret, bytes)
    with pytest.raises(salt.crypt.SaltClientError, match="Nonce verification error"):
        assert master_crypt.loads(ret, nonce="abcde")


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_verify_signature(tmp_path):
    tmp_path.joinpath("foo.pem").write_text(PRIV_KEY.strip())
    tmp_path.joinpath("foo.pub").write_text(PUB_KEY.strip())
    tmp_path.joinpath("bar.pem").write_text(PRIV_KEY2.strip())
    tmp_path.joinpath("bar.pub").write_text(PUB_KEY2.strip())
    msg = b"foo bar"
    sig = salt.crypt.sign_message(str(tmp_path.joinpath("foo.pem")), msg)
    assert salt.crypt.verify_signature(str(tmp_path.joinpath("foo.pub")), msg, sig)


@pytest.mark.skipif(FIPS_TESTRUN, reason="Legacy key can not be loaded in FIPS mode")
def test_verify_signature_bad_sig(tmp_path):
    tmp_path.joinpath("foo.pem").write_text(PRIV_KEY.strip())
    tmp_path.joinpath("foo.pub").write_text(PUB_KEY.strip())
    tmp_path.joinpath("bar.pem").write_text(PRIV_KEY2.strip())
    tmp_path.joinpath("bar.pub").write_text(PUB_KEY2.strip())
    msg = b"foo bar"
    sig = salt.crypt.sign_message(str(tmp_path.joinpath("foo.pem")), msg)
    assert not salt.crypt.verify_signature(str(tmp_path.joinpath("bar.pub")), msg, sig)


def test_read_or_generate_key_string(tmp_path):
    keyfile = tmp_path / ".aes"
    assert not keyfile.exists()
    first_key = salt.crypt.Crypticle.read_key(keyfile)
    assert first_key is None
    assert not keyfile.exists()
    salt.crypt.Crypticle.write_key(keyfile)
    second_key = salt.crypt.Crypticle.read_key(keyfile)
    assert second_key is not None


def test_dropfile_contents(tmp_path, master_opts):
    salt.crypt.dropfile(str(tmp_path), master_opts["user"], master_id=master_opts["id"])
    with salt.utils.files.fopen(str(tmp_path / ".dfn"), "r") as fp:
        assert master_opts["id"] == fp.read()


def test_master_keys_without_cluster_id(tmp_path, master_opts):
    master_opts["pki_dir"] = str(tmp_path)
    assert master_opts["cluster_id"] is None
    assert master_opts["cluster_pki_dir"] is None
    mkeys = salt.crypt.MasterKeys(master_opts)
    expected_master_pub = str(tmp_path / "master.pub")
    expected_master_rsa = str(tmp_path / "master.pem")
    assert expected_master_pub == mkeys.master_pub_path
    assert expected_master_rsa == mkeys.master_rsa_path
    assert mkeys.cluster_pub_path is None
    assert mkeys.cluster_rsa_path is None
    assert mkeys.pub_path == expected_master_pub
    assert mkeys.rsa_path == expected_master_rsa
    assert mkeys.key == mkeys.master_key


def test_master_keys_with_cluster_id(tmp_path, master_opts):
    master_pki_path = tmp_path / "master_pki"
    cluster_pki_path = tmp_path / "cluster_pki"
    # The paths need to exist
    master_pki_path.mkdir()
    cluster_pki_path.mkdir()
    (cluster_pki_path / "peers").mkdir()

    master_opts["pki_dir"] = str(master_pki_path)
    master_opts["cluster_id"] = "cluster1"
    master_opts["cluster_pki_dir"] = str(cluster_pki_path)

    mkeys = salt.crypt.MasterKeys(master_opts)

    expected_master_pub = str(master_pki_path / "master.pub")
    expected_master_rsa = str(master_pki_path / "master.pem")
    expected_cluster_pub = str(cluster_pki_path / "cluster.pub")
    expected_cluster_rsa = str(cluster_pki_path / "cluster.pem")
    assert expected_master_pub == mkeys.master_pub_path
    assert expected_master_rsa == mkeys.master_rsa_path
    assert expected_cluster_pub == mkeys.cluster_pub_path
    assert expected_cluster_rsa == mkeys.cluster_rsa_path
    assert mkeys.pub_path == expected_cluster_pub
    assert mkeys.rsa_path == expected_cluster_rsa
    assert mkeys.key == mkeys.cluster_key


def test_pwdata_decrypt():
    key_string = dedent(
        """
        -----BEGIN RSA PRIVATE KEY-----
        MIIEpQIBAAKCAQEAzhBRyyHa7b63RLE71uKMKgrpulcAJjaIaN68ltXcCvy4w9pi
        Kj+4I3Qp6RvUaHOEmymqyjOMjQc6iwpe0scCFqh3nUk5YYaLZ3WAW0htQVlnesgB
        ZiBg9PBeTQY/LzqtudL6RCng/AX+fbnCsddlIysRxnUoNVMvz0gAmCY2mnTDjcTt
        pyxuk2T0AHSHNCKCalm75L1bWDFF+UzFemf536tBfBUGRWR6jWTij85vvCntxHS/
        HdknaTJ50E7XGVzwBJpCyV4Y2VXuW/3KrCNTqXw+jTmEw0vlcshfDg/vb3IxsUSK
        5KuHalKq/nUIc+F4QCJOl+A10goGdIfYC1/67QIDAQABAoIBAAOP+qoFWtCTZH22
        hq9PWVb8u0+yY1lFxhPyDdaZueUiu1r/coUCdv996Z+TEJgBr0AzdzVpsLtbbaKr
        ujnwoNOdc/vvISPTfKN8P4zUcrcXgZd4z7VhR+vUH/0652q8m/ZDdHorMy2IOP8Z
        cAk9DQ2PmA4TRm+tkX0G5KO8vWLsK921aRMWdsKJyQ0lYxl7M8JWupFsCJFr/U+8
        dAVtwnUiS7RnhBABZ1cfNTHYhXVAh4d+a9y/gZ00a66OGqPxiXfhjjDUZ6fGvWKN
        FlhKWEg6YqIx/H4aNXkLI5Rzzhdx/c2ukNm7+X2veRcAW7bcTwk8wxJxciEP5pBi
        1el9VE0CgYEA/lbzdE2M4yRBvTfYYC6BqZcn+BqtrAUc2h3fEy+p7lwlet0af1id
        gWpYpOJyLc0AUfR616/m2y3PwEH/nMKDSTuU7o/qKNtlHW0nQcnhDCjTUydS3+J/
        JM3dhfgVqi03rjqNcgHA2eOEwcu/OBZtiaC0wqKbuRZRtfGffyoO3ssCgYEAz2iw
        wqu/NkA+MdQIxz/a3Is7gGwoFu6h7O+XU2uN8Y2++jSBw9AzzWj31YCvyjuJPAE+
        gxHm6yOnNoLVn423NtibHejhabzHNIK6UImH99bSTKabsxfF2BX6v982BimU1jwc
        bYykzws37oN/poPb5FTpEiAUrsd2bAMn/1S43icCgYEAulHkY0z0aumCpyUkA8HO
        BvjOtPiGRcAxFLBRXPLL3+vtIQachLHcIJRRf+jLkDXfiCo7W4pm6iWzTbqLkMEG
        AD3/qowPFAM1Hct6uL01efzmYsIp+g0o60NMhvnolRQu+Bm4yM30AyqjdHzYBjSX
        5fyuru8EeSCal1j8aOHcpuUCgYEAhGhDH6Pg59NPYSQJjpm3MMA59hwV473n5Yh2
        xKyO6zwgRT6r8MPDrkhqnwQONT6Yt5PbwnT1Q/t4zhXsJnWkFwFk1U1MSeJYEa+7
        HZsPECs2CfT6xPRSO0ac00y+AmUdPT8WruDwfbSdukh8f2MCR9vlBsswKPvxH7dM
        G3aMplUCgYEAmMFgB/6Ox4OsQPPC6g4G+Ezytkc4iVkMEcjiVWzEsYATITjq3weO
        /XDGBYJoBhYwWPi9oBufFc/2pNtWy1FKKXPuVyXQATdA0mfEPbtsHjMFQNZbeKnm
        0na/SysSDCK3P+9ijlbjqLjMmPEmhJxGWTJ7khnTTkfre7/w9ZxJxi8=
        -----END RSA PRIVATE KEY-----"""
    )

    pwdata = (
        b"V\x80+b\xca\x06M\xb6\x12\xc6\xe8\xf2\xb5\xbb\xd8m\xc0\x97\x9a\xeb\xb9q\x19\xc3"
        b'\xcdi\xb84\x90\xaf\x12kT\xe2@u\xd6\xe8T\x89\xa3\xc7\xb2Y\xd1N\x00\xa9\xc0"\xbe'
        b"\xed\xb1\xc3\xb7^\xbf\xbd\x8b\x13\xd3/L\x1b\xa1`\xe2\xea\x03\x98\x82\xf3uS&|"
        b'\xe5\xd8J\xce\xfc\x97\x8d\x0b\x949\xc0\xbd^\xef\xc6\xfd\xce\xbb\x1e\xd0"(m\xe1'
        b"\x95\xfb\xc8/\x07\x93\xb8\xda\x8f\x99\xfe\xdc\xd5\xcb\xdb\xb2\xf11M\xdbD\xcf"
        b"\x95\x13p\r\xa4\x1c{\xd5\xdb\xc7\xe5\xaf\x95F\x97\xa9\x00p~\xb5\xec\xa4.\xd0"
        b"\xa4\xb4\xf4f\xcds,Y/\xa1:WF\xb8\xc7\x07\xaa\x0b<'~\x1b$D9\xd4\x8d\xf0x\xc5"
        b"\xee\xa8:\xe6\x00\x10\xc5i\x11\xc7]C8\x05l\x8b\x9b\xc3\x83e\xf7y\xadi:0\xb4R"
        b"\x1a(\x04&yL8\x19s\n\x11\x81\xfd?\xfb2\x80Ll\xa1\xdc\xc9\xb6P\xca\x8d'\x11\xc1"
        b"\x07\xa5\xa1\x058\xc7\xce\xbeb\x92\xbf\x0bL\xec\xdf\xc3M\x83\xfb$\xec\xd5\xf9"
    )
    assert salt.crypt.pwdata_decrypt(key_string, pwdata) == "1234"



"""
Support for pam
"""

import logging
import os

import salt.utils.files

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pam"


def __virtual__():
    """
    Set the virtual name for the module
    """
    return __virtualname__


def _parse(contents=None, file_name=None):
    """
    Parse a standard pam config file
    """
    if contents:
        pass
    elif file_name and os.path.exists(file_name):
        with salt.utils.files.fopen(file_name, "r") as ifile:
            contents = salt.utils.stringutils.to_unicode(ifile.read())
    else:
        log.error('File "%s" does not exist', file_name)
        return False

    rules = []
    for line in contents.splitlines():
        if not line:
            continue
        if line.startswith("#"):
            continue
        control_flag = ""
        module = ""
        arguments = []
        comps = line.split()
        interface = comps[0]
        position = 1
        if comps[1].startswith("["):
            control_flag = comps[1].replace("[", "")
            for part in comps[2:]:
                position += 1
                if part.endswith("]"):
                    control_flag += " {}".format(part.replace("]", ""))
                    position += 1
                    break
                else:
                    control_flag += f" {part}"
        else:
            control_flag = comps[1]
            position += 1
        module = comps[position]
        if len(comps) > position:
            position += 1
            arguments = comps[position:]
        rules.append(
            {
                "interface": interface,
                "control_flag": control_flag,
                "module": module,
                "arguments": arguments,
            }
        )
    return rules


def read_file(file_name):
    """
    This is just a test function, to make sure parsing works

    CLI Example:

    .. code-block:: bash

        salt '*' pam.read_file /etc/pam.d/login
    """
    return _parse(file_name=file_name)



import pytest

import salt.channel.server as server


@pytest.fixture
def key_data():
    return [
        "-----BEGIN PUBLIC KEY-----",
        "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoe5QSDYRWKyknbVyRrIj",
        "rm1ht5HgKzAVUber0x54+b/UgxTd1cqI6I+eDlx53LqZSH3G8Rd5cUh8LHoGedSa",
        "E62vEiLAjgXa+RdgcGiQpYS8+Z2RvQJ8oIcZgO+2AzgBRHboNWHTYRRmJXCd3dKs",
        "9tcwK6wxChR06HzGqaOTixAuQlegWbOTU+X4dXIbW7AnuQBt9MCib7SxHlscrqcS",
        "cBrRvq51YP6cxPm/rZJdBqZhVrlghBvIpa45NApP5PherGi4AbEGYte4l+gC+fOA",
        "osEBis1V27djPpIyQS4qk3XAPQg6CYQMDltHqA4Fdo0Nt7SMScxJhfH0r6zmBFAe",
        "BQIDAQAB",
        "-----END PUBLIC KEY-----",
    ]


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys(key_data, linesep):
    src_key = linesep.join(key_data)
    tgt_key = "\n".join(key_data)
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys_newline_src(key_data, linesep):
    src_key = linesep.join(key_data) + linesep
    tgt_key = "\n".join(key_data)
    assert src_key.endswith(linesep)
    assert not tgt_key.endswith("\n")
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True


@pytest.mark.parametrize("linesep", ["\r\n", "\r", "\n"])
def test_compare_keys_newline_tgt(key_data, linesep):
    src_key = linesep.join(key_data)
    tgt_key = "\n".join(key_data) + "\n"
    assert not src_key.endswith(linesep)
    assert tgt_key.endswith("\n")
    assert server.ReqServerChannel.compare_keys(src_key, tgt_key) is True



import pytest

from tests.support.case import ModuleCase


@pytest.mark.requires_salt_modules("rabbitmq")
@pytest.mark.windows_whitelisted
@pytest.mark.skip_if_not_root
class RabbitModuleTest(ModuleCase):
    """
    Validates the rabbitmqctl functions.
    To run these tests, you will need to be able to access the rabbitmqctl
    commands.
    """

    def test_user_exists(self):
        """
        Find out whether a user exists.
        """
        ret = self.run_function("rabbitmq.user_exists", ["null_user"])
        self.assertEqual(ret, False)



"""
salt-ssh testing
"""

import pathlib
import shutil

import pytest

import salt.utils.files
import salt.utils.yaml
from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(autouse=True)
def thin_dir(salt_ssh_cli):
    try:
        yield
    finally:
        ret = salt_ssh_cli.run("config.get", "thin_dir")
        assert ret.returncode == 0
        thin_dir_path = ret.data
        shutil.rmtree(thin_dir_path, ignore_errors=True)


@pytest.fixture(scope="module")
def invalid_json_exe_mod(salt_run_cli, base_env_state_tree_root_dir):
    module_contents = r"""
import os
import sys


def __virtual__():
    return "whoops"


def test():
    data = '{\n  "local": {\n    "whoops": "hrhrhr"\n  }\n}'
    ctr = 0
    for line in data.splitlines():
        sys.stdout.write(line)
        if ctr == 3:
            print("Warning: Chaos is not a letter")
        ctr += 1
    sys.stdout.flush()
    os._exit(0)
"""
    module_dir = base_env_state_tree_root_dir / "_modules"
    module_tempfile = pytest.helpers.temp_file("whoops.py", module_contents, module_dir)
    try:
        with module_tempfile:
            ret = salt_run_cli.run("saltutil.sync_modules")
            assert ret.returncode == 0
            assert "modules.whoops" in ret.data
            yield
    finally:
        ret = salt_run_cli.run("saltutil.sync_modules")
        assert ret.returncode == 0


@pytest.fixture(scope="module")
def invalid_return_exe_mod(salt_run_cli, base_env_state_tree_root_dir):
    module_contents = r"""
import json
import os
import sys


def __virtual__():
    return "whoopsiedoodle"


def test(wrapped=True):
    data = "Chaos is a ladder though"
    if wrapped:
        data = {"local": {"no_return_key_present": data}}
    else:
        data = {"no_local_key_present": data}

    print(json.dumps(data))
    sys.stdout.flush()
    os._exit(0)
"""
    module_dir = base_env_state_tree_root_dir / "_modules"
    module_tempfile = pytest.helpers.temp_file(
        "whoopsiedoodle.py", module_contents, module_dir
    )
    try:
        with module_tempfile:
            ret = salt_run_cli.run("saltutil.sync_modules")
            assert ret.returncode == 0
            assert "modules.whoopsiedoodle" in ret.data
            yield
    finally:
        ret = salt_run_cli.run("saltutil.sync_modules")
        assert ret.returncode == 0


@pytest.fixture(scope="module")
def remote_exception_wrap_mod(salt_master):
    module_contents = r"""
def __virtual__():
    return "check_exception"


def failure():
    # This should raise an exception
    ret = __salt__["disk.usage"]("c")
    return f"Probably got garbage: {ret}"
"""
    module_dir = pathlib.Path(salt_master.config["extension_modules"]) / "wrapper"
    module_tempfile = pytest.helpers.temp_file(
        "check_exception.py", module_contents, module_dir
    )
    with module_tempfile:
        yield


@pytest.fixture(scope="module")
def remote_parsing_failure_wrap_mod(salt_master, invalid_json_exe_mod):
    module_contents = r"""
def __virtual__():
    return "check_parsing"


def failure(mod):
    # This should raise an exception
    ret = __salt__[f"{mod}.test"]()
    return f"Probably got garbage: {ret}"
"""
    module_dir = pathlib.Path(salt_master.config["extension_modules"]) / "wrapper"
    module_tempfile = pytest.helpers.temp_file(
        "check_parsing.py", module_contents, module_dir
    )
    with module_tempfile:
        yield


def test_ping(salt_ssh_cli):
    """
    Test a simple ping
    """
    ret = salt_ssh_cli.run("test.ping")
    assert ret.returncode == 0
    assert ret.data is True


def test_thin_dir(salt_ssh_cli):
    """
    test to make sure thin_dir is created
    and salt-call file is included
    """
    ret = salt_ssh_cli.run("config.get", "thin_dir")
    assert ret.returncode == 0
    thin_dir = pathlib.Path(ret.data)
    assert thin_dir.is_dir()
    assert thin_dir.joinpath("salt-call").exists()
    assert thin_dir.joinpath("running_data").exists()


def test_wipe(salt_ssh_cli):
    """
    Ensure --wipe is respected by the state module wrapper
    issue #61083
    """
    ret = salt_ssh_cli.run("config.get", "thin_dir")
    assert ret.returncode == 0
    thin_dir = pathlib.Path(ret.data)
    assert thin_dir.exists()
    # only few modules (state and cp) will actually respect --wipe
    # (see commit #8a414d53284ec04940540ebd823306ab5119e105)
    salt_ssh_cli.run("--wipe", "state.apply")
    assert not thin_dir.exists()


def test_set_path(salt_ssh_cli, tmp_path, salt_ssh_roster_file):
    """
    test setting the path env variable
    """
    path = "/pathdoesnotexist/"
    roster_file = tmp_path / "roster-set-path"
    with salt.utils.files.fopen(salt_ssh_roster_file) as rfh:
        roster_data = salt.utils.yaml.safe_load(rfh)
        roster_data["localhost"].update(
            {
                "set_path": f"$PATH:/usr/local/bin/:{path}",
            }
        )
    with salt.utils.files.fopen(roster_file, "w") as wfh:
        salt.utils.yaml.safe_dump(roster_data, wfh)

    ret = salt_ssh_cli.run(f"--roster-file={roster_file}", "environ.get", "PATH")
    assert ret.returncode == 0
    assert path in ret.data


def test_tty(salt_ssh_cli, tmp_path, salt_ssh_roster_file):
    """
    test using tty
    """
    roster_file = tmp_path / "roster-tty"
    with salt.utils.files.fopen(salt_ssh_roster_file) as rfh:
        roster_data = salt.utils.yaml.safe_load(rfh)
        roster_data["localhost"].update({"tty": True})
    with salt.utils.files.fopen(roster_file, "w") as wfh:
        salt.utils.yaml.safe_dump(roster_data, wfh)
    ret = salt_ssh_cli.run(f"--roster-file={roster_file}", "test.ping")
    assert ret.returncode == 0
    assert ret.data is True


def test_retcode_exe_run_fail(salt_ssh_cli):
    """
    Verify salt-ssh passes through the retcode it receives.
    """
    ret = salt_ssh_cli.run("file.touch", "/tmp/non/ex/is/tent")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    # This should be the exact output, but some other warnings
    # might be printed to stderr.
    assert "Error running 'file.touch': No such file or directory" in ret.data["stderr"]


def test_retcode_exe_run_exception(salt_ssh_cli):
    """
    Verify salt-ssh passes through the retcode it receives
    when an exception is thrown. (Ref #50727)
    """
    ret = salt_ssh_cli.run("salttest.jinja_error")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert ret.data["stderr"].endswith("Exception: hehehe")


@pytest.mark.usefixtures("invalid_json_exe_mod")
def test_retcode_json_decode_error(salt_ssh_cli):
    """
    Verify salt-ssh exits with a non-zero exit code when
    it cannot decode the output of a command.
    """
    ret = salt_ssh_cli.run("whoops.test")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert (
        ret.data["stdout"]
        == '{  "local": {    "whoops": "hrhrhr"  }Warning: Chaos is not a letter\n}'
    )
    assert ret.data["_error"] == "Failed to return clean data"
    assert ret.data["retcode"] == 0


@pytest.mark.usefixtures("invalid_return_exe_mod")
def test_retcode_invalid_return(salt_ssh_cli):
    """
    Verify salt-ssh exits with a non-zero exit code when
    the decoded command output is invalid.
    """
    ret = salt_ssh_cli.run("whoopsiedoodle.test", "false")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert ret.data["stdout"] == '{"no_local_key_present": "Chaos is a ladder though"}'
    assert ret.data["_error"] == "Return dict was malformed"
    assert ret.data["retcode"] == 0
    assert ret.data["parsed"] == {"no_local_key_present": "Chaos is a ladder though"}


@pytest.mark.usefixtures("remote_exception_wrap_mod")
def test_wrapper_unwrapped_command_exception(salt_ssh_cli):
    """
    Verify salt-ssh does not return unexpected exception output to wrapper modules.
    """
    ret = salt_ssh_cli.run("check_exception.failure")
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, dict)
    assert ret.data
    assert "Probably got garbage" not in ret.data["stderr"]
    assert (
        "Error running 'disk.usage': Invalid flag passed to disk.usage"
        in ret.data["stderr"]
    )


@pytest.mark.usefixtures("remote_parsing_failure_wrap_mod", "invalid_json_exe_mod")
def test_wrapper_unwrapped_command_parsing_failure(salt_ssh_cli):
    """
    Verify salt-ssh does not return unexpected unparsable output to wrapper modules.
    """
    ret = salt_ssh_cli.run("check_parsing.failure", "whoops")
    assert ret.returncode == EX_AGGREGATE
    assert ret.data
    assert "Probably got garbage" not in ret.data["stderr"]
    assert isinstance(ret.data, dict)
    assert ret.data["_error"] == "Failed to return clean data"
    assert ret.data["retcode"] == 0
    assert (
        ret.data["stdout"]
        == '{  "local": {    "whoops": "hrhrhr"  }Warning: Chaos is not a letter\n}'
    )


@pytest.mark.usefixtures("remote_parsing_failure_wrap_mod", "invalid_return_exe_mod")
def test_wrapper_unwrapped_command_invalid_return(salt_ssh_cli):
    """
    Verify salt-ssh does not return unexpected unparsable output to wrapper modules.
    """
    ret = salt_ssh_cli.run("check_parsing.failure", "whoopsiedoodle")
    assert ret.returncode == EX_AGGREGATE
    assert ret.data
    assert "Probably got garbage" not in ret.data
    assert isinstance(ret.data, dict)
    assert ret.data["_error"] == "Return dict was malformed"
    assert ret.data["retcode"] == 0
    assert (
        ret.data["stdout"]
        == '{"local": {"no_return_key_present": "Chaos is a ladder though"}}'
    )
    assert ret.data["parsed"] == {
        "local": {"no_return_key_present": "Chaos is a ladder though"}
    }


@pytest.fixture(scope="module")
def utils_dependent_module(salt_run_cli, salt_master):
    module_contents = r"""
import customutilsmodule


def __virtual__():
    return "utilsync"


def test():
    return customutilsmodule.test()
"""
    utils_contents = r"""
def test():
    return "success"
"""
    module_tempfile = salt_master.state_tree.base.temp_file(
        "_modules/utilsync.py", contents=module_contents
    )
    util_tempfile = salt_master.state_tree.base.temp_file(
        "_utils/customutilsmodule.py", contents=utils_contents
    )
    with module_tempfile, util_tempfile:
        yield


@pytest.mark.usefixtures("utils_dependent_module")
def test_custom_utils_are_present_on_target(salt_ssh_cli):
    ret = salt_ssh_cli.run("utilsync.test")
    assert ret.returncode == 0
    assert ret.data == "success"
